# -*- coding: utf-8 -*-
# Author: allen
# Copyright (c) 2020 RuiCare All rights reserved.
import collections
import os
import pydicom
from util import fileutil
from pydicom import Dataset


def get_dcm_data(file, raise_error=False):
    dcm_data = None
    try:
        dcm_data_tmp = pydicom.dcmread(file, force=True)
        if dcm_data_tmp is None or (dcm_data_tmp and dcm_data_tmp.get("Modality") is None):
            raise RuntimeError('不能读取非DICOM类型文件')
        else:
            dcm_data = dcm_data_tmp
    except Exception as e:
        if raise_error: raise e
    return dcm_data


def get_file_meta_dataset(sop_instance_uid):
    file_meta = Dataset()
    file_meta.FileMetaInformationGroupLength = 146 + sop_instance_uid.__len__()
    file_meta.FileMetaInformationVersion = b"\x00\x01"
    file_meta.MediaStorageSOPClassUID = "1.2.840.10008.5.1.4.1.1.481.3"
    file_meta.MediaStorageSOPInstanceUID = sop_instance_uid
    file_meta.TransferSyntaxUID = "1.2.840.10008.1.2"
    file_meta.ImplementationClassUID = "1.2.276.0.7230010.3.0.3.5.3"
    return file_meta


def get_ct_info(ct_dir):
    ct_paths = fileutil.list_all_file_path(ct_dir)
    if not ct_paths or ct_paths.__len__() == 0: return None
    ct_info = collections.OrderedDict()
    ct_data = get_dcm_data(ct_paths[0])
    ct_info['Rows'] = ct_data.Rows
    ct_info['Columns'] = ct_data.Columns
    ct_info['PixelSpacing'] = list(ct_data.PixelSpacing)
    # ct_info['WindowCenter'] = list(ct_data.WindowCenter) if isinstance(ct_data.WindowCenter, MultiValue) else str(
    #     ct_data.WindowCenter)
    # ct_info['WindowWidth'] = list(ct_data.WindowWidth) if isinstance(ct_data.WindowWidth, MultiValue) else str(
    #     ct_data.WindowWidth)
    ct_info['RescaleIntercept'] = ct_data.RescaleIntercept if hasattr(ct_data, 'RescaleIntercept') else None
    ct_info['RescaleSlope'] = ct_data.RescaleSlope if hasattr(ct_data, 'RescaleSlope') else None
    ct_info['SliceThickness'] = ct_data.SliceThickness
    ct_info['PatientPosition'] = ct_data.PatientPosition
    ct_info['FrameOfReferenceUID'] = ct_data.FrameOfReferenceUID
    ct_info['ImagePositionPatient'] = list(ct_data.ImagePositionPatient)
    ct_info['ImageOrientationPatient'] = list(ct_data.ImageOrientationPatient)
    index_z_map = collections.OrderedDict()
    index_ct_uid_map = collections.OrderedDict()
    for ct_path in ct_paths:
        dcm_data = get_dcm_data(ct_path)
        z = 0.00 if float(dcm_data.ImagePositionPatient[2]) == 0 else float(dcm_data.ImagePositionPatient[2])
        index_z_map[dcm_data.InstanceNumber] = '{:.2f}'.format(z)
        index_ct_uid_map[dcm_data.InstanceNumber] = dcm_data.SOPInstanceUID
    index_z_map = sorted(index_z_map.items(), key=lambda i: int(i[0]))
    index_ct_uid_map = sorted(index_ct_uid_map.items(), key=lambda i: int(i[0]))
    ct_info['index_z_map'] = collections.OrderedDict(index_z_map)  # 此处collections.OrderedDict不可去掉，用于将数组转成字典
    ct_info['index_ct_uid_map'] = collections.OrderedDict(index_ct_uid_map)
    return ct_info


def ct_to_nii(ct_dir, out_nii_path):
    # import dicom2nifti
    # reorient_nifti = False
    # dicom2nifti.settings.enable_pydicom_read_force()
    # dicom2nifti.dicom_series_to_nifti(ct_dir, out_nii_path, reorient_nifti)
    convert_ct_cmd = 'plastimatch convert --input "{0}" --output-img "{1}"'
    convert_ct_cmd = convert_ct_cmd.format(ct_dir, out_nii_path)
    shell = os.popen(convert_ct_cmd)
    convert_ct_result = shell.read()
    convert_ct_exitcode = shell.close()
    if convert_ct_exitcode is not None and convert_ct_exitcode != 0:
        raise RuntimeError('Convert CT: %s catch error: %s' % (ct_dir, convert_ct_result))


def rs_to_nii(plastimatch_bin, ct_dir, rs_path, out_nii_dir):
    convert_rs_cmd = '{0} convert --input "{1}" --prefix-format nii.gz --referenced-ct "{2}" --output-prefix "{3}"'
    convert_rs_cmd = convert_rs_cmd.format(plastimatch_bin, rs_path, ct_dir, out_nii_dir)
    shell = os.popen(convert_rs_cmd)
    convert_rs_result = shell.read()
    convert_rs_exitcode = shell.close()
    if convert_rs_exitcode is not None and convert_rs_exitcode != 0:
        raise RuntimeError('Convert RS: %s catch error: %s' % (rs_path, convert_rs_result))


def get_rs_dcm(dcm_pat_dir):
    dcm_paths = fileutil.list_all_file_path(dcm_pat_dir)
    rs_dcm_datas, rs_dcm_paths = [], []
    for dcm_path in dcm_paths:
        dcm_data = get_dcm_data(dcm_path)
        md = dcm_data.get("Modality")
        if md is not None and md == 'RTSTRUCT':
            rs_dcm_datas.append(dcm_data)
            rs_dcm_paths.append(dcm_path)
    return rs_dcm_datas, rs_dcm_paths


def get_newest_rs(dcm_pat_dir):
    _, rs_dcm_paths = get_rs_dcm(dcm_pat_dir)
    rs_dcm_paths.sort(key=lambda x: os.path.getmtime(x))
    return (get_dcm_data(rs_dcm_paths[0]), rs_dcm_paths[0]) if rs_dcm_paths else (None, None)
