import os
import cv2
import nibabel
import numpy as np
from PIL import Image
import SimpleITK as sitk
from nibabel.pydicom_compat import pydicom


def dispersed(min=-1024, max=2048):
    def __dispersed__(x):
        if min and max:
            if min: x[x < min] = min
            if max: x[x > max] = max
        __min__, __max__ = min if min else x.min(), max if max else x.max()
        return (x - __min__) / (__max__ - __min__) if __max__ - __min__ != 0 else x

    return __dispersed__


def resize_crop_scale(img, first_resize=None, crop_size=None, last_resize=None, scale=1):
    original_img_dtype = img.dtype
    if original_img_dtype == np.int32 or original_img_dtype == np.int64: img = img.astype(np.int16)
    if first_resize: img = np.array(Image.fromarray(img).resize((first_resize[1], first_resize[0]), resample=0))
    if crop_size: img = img[crop_size[0]:crop_size[1], crop_size[2]:crop_size[3]]
    if last_resize: img = np.array(Image.fromarray(img).resize((last_resize[1], last_resize[0]), resample=0))
    new_rows, new_cols = int(scale * img.shape[0]), int(scale * img.shape[1])
    assert new_rows > 0 and new_cols > 0, 'Scale is too small'
    if scale != 1: img = np.array(Image.fromarray(img).resize((new_rows, new_cols), resample=0))
    return img.astype(original_img_dtype)


def normalize(img, range=255, enable_clahe=True, clahe=2.55):
    img = np.float32(img)
    if img.min() == 0.0 and img.max() == 0.0: return np.zeros(shape=img.shape, dtype=np.uint8)
    img = range * (img - img.min()) / (img.max() - img.min())
    img = np.uint8(np.around(img))
    if enable_clahe: img = cv2.createCLAHE(clahe).apply(img)
    return img


def get_normalizer(min=None, max=None):
    def __dispersed__(x):
        if min and max:
            if min: x[x < min] = min
            if max: x[x > max] = max
        __min__, __max__ = min if min else x.min(), max if max else x.max()
        return (x - __min__) / (__max__ - __min__) if __max__ - __min__ != 0 else x

    return __dispersed__


def load_nii(path, dtype=None, zfirst=True):
    nii = sitk.ReadImage(path)
    img = sitk.GetArrayFromImage(nii)
    if not zfirst: img = img.transpose((1, 2, 0))
    if dtype: img = img.astype(dtype)
    return img


def write_nii(img, path, spacing=None, origin=None):
    out = sitk.GetImageFromArray(img)
    if spacing: out.SetSpacing(spacing)
    if origin: out.SetOrigin(origin)
    sitk.WriteImage(out, path)


def resample_with_spacing(itk_img, new_spacing):
    rs = sitk.ResampleImageFilter()
    rs.SetInterpolator(sitk.sitkLinear)
    rs.SetOutputDirection(itk_img.GetDirection())
    rs.SetOutputOrigin(itk_img.GetOrigin())
    new_spacing = np.array(new_spacing, float)
    size = itk_img.GetSize() / new_spacing * itk_img.GetSpacing()
    rs.SetSize(size.astype(np.int).tolist())
    rs.SetOutputSpacing(new_spacing)
    return rs.Execute(itk_img)
